#
# Copyright (c) 2001 by Jim Menard <jimm@io.com>
#
# Released under the same license as Ruby. See
# http://www.ruby-lang.org/en/LICENSE.txt.
#

module NQXML

    VERSION_MAJOR = 1
    VERSION_MINOR = 1
    VERSION_TWEAK = 3
    Version = "#{VERSION_MAJOR}.#{VERSION_MINOR}.#{VERSION_TWEAK}"
    Copyright = 'Copyright (c) 2001 by Jim Menard <jimm@io.com>'

end
